{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "7320dcbc",
   "metadata": {},
   "source": [
    "# Thread Safety\n",
    "\n",
    "[![Twitter Handle](https://img.shields.io/badge/Twitter-@gaohongnan-blue?style=social&logo=twitter)](https://twitter.com/gaohongnan)\n",
    "[![LinkedIn Profile](https://img.shields.io/badge/@gaohongnan-blue?style=social&logo=linkedin)](https://linkedin.com/in/gao-hongnan)\n",
    "[![GitHub Profile](https://img.shields.io/badge/GitHub-gao--hongnan-lightgrey?style=social&logo=github)](https://github.com/gao-hongnan)\n",
    "![Tag](https://img.shields.io/badge/Tag-Brain_Dump-red)\n",
    "![Tag](https://img.shields.io/badge/Level-Beginner-green)\n",
    "[![Code](https://img.shields.io/badge/View-Code-blue?style=flat-square&logo=github)](https://github.com/gao-hongnan/omniverse/blob/8ddcd6a345925e7fd550b74ce4577a0e2807aa5f/omnixamples/software_engineering/concurrency_parallelism_asynchronous/race_condition.py)\n",
    "\n",
    "```{contents}\n",
    ":local:\n",
    "```\n",
    "\n",
    "With pre-emptive multitasking, the OS can interrupt a thread at any time, even\n",
    "in the middle of executing a single Python statement. This can lead to issues\n",
    "like **race conditions**, where the program's behavior depends on the\n",
    "unpredictable timing of threads.\n",
    "\n",
    "Consider the below code:\n",
    "\n",
    "```python\n",
    "\"\"\"With reference to effective python book chapter 54.\n",
    "Ref: https://github.com/bslatkin/effectivepython/blob/master/example_code/item_54.py\n",
    "\"\"\"\n",
    "import logging\n",
    "import threading\n",
    "from threading import Barrier\n",
    "from typing import List\n",
    "\n",
    "logging.basicConfig(level=logging.DEBUG, format=\"%(asctime)s - %(levelname)s - %(message)s\")\n",
    "\n",
    "NUM_THREADS = 5\n",
    "BARRIER = Barrier(NUM_THREADS)\n",
    "\n",
    "\n",
    "class Counter:\n",
    "    def __init__(self) -> None:\n",
    "        self.count = 0\n",
    "\n",
    "    def increment(self, offset: int) -> None:\n",
    "        self.count += offset\n",
    "\n",
    "def worker(thread_index: int, total_iterations: int, counter: Counter) -> None:\n",
    "    \"\"\"The barrier is used to synchronize the threads so that they all start counting\n",
    "    at the same time. This makes it easier to get a race condition since we wait for\n",
    "    the other threads to start else in the loop we always have an order that the\n",
    "    first thread likely starts first and then the second and so on.\n",
    "    \"\"\"\n",
    "    BARRIER.wait()\n",
    "    logging.debug(\"Thread %s, starting\", thread_index)\n",
    "    for _ in range(total_iterations):\n",
    "        counter.increment(1)\n",
    "\n",
    "\n",
    "def thread_unsafe(total_iterations: int) -> None:\n",
    "    counter = Counter()\n",
    "\n",
    "    threads: List[threading.Thread] = []\n",
    "    for index in range(NUM_THREADS):\n",
    "        thread = threading.Thread(target=worker, args=(index, total_iterations, counter))\n",
    "        threads.append(thread)\n",
    "\n",
    "    for thread in threads:\n",
    "        thread.start()\n",
    "\n",
    "    for thread in threads:\n",
    "        thread.join()\n",
    "\n",
    "    expected = total_iterations * NUM_THREADS\n",
    "    found = counter.count\n",
    "    logging.info(\"Counter should be %s, got %s\", expected, found)\n",
    "\n",
    "\n",
    "if __name__ == \"__main__\":\n",
    "    total_iterations = 10**6\n",
    "\n",
    "    thread_unsafe(total_iterations)\n",
    "```\n",
    "\n",
    "There are a total of 5 threads, each incrementing the counter by 1 for $10^6$\n",
    "times. So ideally the **_expected output_** should be\n",
    "$5 \\times 10^6 = 5,000,000$. However, it could be less, for example, we may get\n",
    "$2,000,000$ or $3,000,000$ sometimes.\n",
    "\n",
    "## Why Does This Happen?\n",
    "\n",
    "The line `counter.increment(1)` seems atomic but is actually a shorthand for:\n",
    "\n",
    "1. **Read:** Retrieve the current value of `counter`:\n",
    "   `current_value = counter.count`\n",
    "2. **Add:** Increment the value by 1: `new_value = current_value + 1`\n",
    "3. **Write:** Store the new value back to `counter`: `counter.count = new_value`\n",
    "\n",
    "If the OS pre-empts a thread **after** reading but **before** writing, another\n",
    "thread might read the same initial value, leading to lost updates.\n",
    "\n",
    "Let's pan out a scenario where this can happen:\n",
    "\n",
    "1. Thread 1: `thread_1_current_value = counter.count = 0`\n",
    "2. Thread 2: `thread_2_current_value = counter.count = 0`. This is called a\n",
    "   _context switch_.\n",
    "3. Thread 2: `thread_2_new_value = thread_2_current_value + 1 = 1`\n",
    "4. Thread 2: `counter.count = thread_2_new_value = 1`.\n",
    "5. Thread 1: `thread_1_new_value = thread_1_current_value + 1 = 1`\n",
    "6. Thread 1: `counter.count = thread_1_new_value = 1`.\n",
    "\n",
    "So we see thread 2 interrupted thread 1 after thread 1 read the value of\n",
    "`counter` but before it could write the new value back. Then even though the\n",
    "counter should be 2, it is 1 because thread 1 overwrote it.\n",
    "\n",
    "## Lock\n",
    "\n",
    "Python's threading module offers a suite of tools to prevent issues like data\n",
    "races and data structure corruption in multithreaded environments. Among these,\n",
    "the `Lock` class stands out as a particularly useful and straightforward option.\n",
    "It implements a mutual-exclusion lock, commonly known as a mutex.\n",
    "\n",
    "By incorporating a `Lock` into the `Counter` class, you can safeguard its\n",
    "current value from concurrent access by multiple threads. The lock ensures that\n",
    "only one thread can access the protected data at any given moment. To manage the\n",
    "lock efficiently, you can utilize Python's with statement, which handles both\n",
    "the acquisition and release of the lock automatically.\n",
    "\n",
    "```python\n",
    "\"\"\"With reference to effective python book chapter 54.\n",
    "Ref: https://github.com/bslatkin/effectivepython/blob/master/example_code/item_54.py\n",
    "\"\"\"\n",
    "import logging\n",
    "import threading\n",
    "from threading import Barrier\n",
    "from typing import List\n",
    "\n",
    "logging.basicConfig(level=logging.DEBUG, format=\"%(asctime)s - %(levelname)s - %(message)s\")\n",
    "\n",
    "NUM_THREADS = 5\n",
    "BARRIER = Barrier(NUM_THREADS)\n",
    "\n",
    "class CounterLock:\n",
    "    def __init__(self) -> None:\n",
    "        self.count = 0\n",
    "        self.lock = threading.Lock()\n",
    "\n",
    "    def increment(self, offset: int) -> None:\n",
    "        with self.lock:\n",
    "            self.count += offset\n",
    "\n",
    "\n",
    "def worker(thread_index: int, total_iterations: int, counter: Counter) -> None:\n",
    "    \"\"\"The barrier is used to synchronize the threads so that they all start counting\n",
    "    at the same time. This makes it easier to get a race condition since we wait for\n",
    "    the other threads to start else in the loop we always have an order that the\n",
    "    first thread likely starts first and then the second and so on.\n",
    "    \"\"\"\n",
    "    BARRIER.wait()\n",
    "    logging.debug(\"Thread %s, starting\", thread_index)\n",
    "    for _ in range(total_iterations):\n",
    "        counter.increment(1)\n",
    "\n",
    "\n",
    "def thread_safe(total_iterations: int) -> None:\n",
    "    counter = CounterLock()\n",
    "\n",
    "    threads: List[threading.Thread] = []\n",
    "    for index in range(NUM_THREADS):\n",
    "        thread = threading.Thread(target=worker, args=(index, total_iterations, counter))\n",
    "        threads.append(thread)\n",
    "\n",
    "    for thread in threads:\n",
    "        thread.start()\n",
    "\n",
    "    for thread in threads:\n",
    "        thread.join()\n",
    "\n",
    "    expected = total_iterations * NUM_THREADS\n",
    "    found = counter.count\n",
    "\n",
    "    logging.info(\"Counter should be %s, got %s\", expected, found)\n",
    "\n",
    "\n",
    "if __name__ == \"__main__\":\n",
    "    total_iterations = 10**6\n",
    "\n",
    "    thread_safe(total_iterations)\n",
    "```\n",
    "\n",
    "## References And Further Readings\n",
    "\n",
    "-   [Chapter 54: Use Lock to Prevent Data Races in Threads - Effective Python](https://github.com/bslatkin/effectivepython/blob/master/example_code/item_54.py)"
   ]
  }
 ],
 "metadata": {
  "jupytext": {
   "cell_metadata_filter": "-all",
   "formats": "md:myst",
   "text_representation": {
    "extension": ".md",
    "format_name": "myst",
    "format_version": 0.13,
    "jupytext_version": "1.11.5"
   }
  },
  "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "mystnb": {
   "number_source_lines": true
  },
  "source_map": [
   16
  ]
 },
 "nbformat": 4,
 "nbformat_minor": 5
}